/********************************************************************\
 * Transaction.h -- defines transaction for xacc (X-Accountant)     *
 * Copyright (C) 1997 Robin D. Clark                                *
 * Copyright (C) 1997 Linas Vepstas                                 *
 *                                                                  *
 * This program is free software; you can redistribute it and/or    *
 * modify it under the terms of the GNU General Public License as   *
 * published by the Free Software Foundation; either version 2 of   *
 * the License, or (at your option) any later version.              *
 *                                                                  *
 * This program is distributed in the hope that it will be useful,  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    *
 * GNU General Public License for more details.                     *
 *                                                                  *
 * You should have received a copy of the GNU General Public License*
 * along with this program; if not, write to the Free Software      *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.        *
 *                                                                  *
 *   Author: Rob Clark                                              *
 * Internet: rclark@cs.hmc.edu                                      *
 *  Address: 609 8th Street                                         *
 *           Huntington Beach, CA 92648-4632                        *
\********************************************************************/

#ifndef __XACC_TRANSACTION_H__
#define __XACC_TRANSACTION_H__

#include "config.h"
#include "date.h"   /* for Date */

/* Values for the reconciled field in Transaction: */
#define CREC 'c'              /* The transaction has been cleared        */
#define YREC 'y'              /* The transaction has been reconciled     */
#define FREC 'f'              /* frozen into accounting period           */
#define NREC 'n'              /* not reconciled or cleared               */

/** STRUCTS *********************************************************/
/* The debit & credit pointers are used to implement a double-entry 
 * accounting system.  Basically, the idea with double entry is that
 * there is always an account that is debited, and another that is
 * credited.  These two pointers identify the two accounts. 
 */

/* A split transaction is one which shows up as a credit (or debit) in
 * one account, and peices of it show up as debits (or credits) in other
 * accounts.  Thus, a single credit-card transaction might be split
 * between "dining", "tips" and "taxes" categories.
 */

typedef struct _split {
  struct _account *acc;     /* back-pointer to debited/credited account   */
  struct _transaction *parent; /* parent of split                         */
  char  * memo;
  char  * action;            /* Buy, Sell, Div, etc.                      */
  char    reconciled;
  double  damount;           /* num-shares; if > 0.0, deposit, else paymt */
  double  share_price;       /* the share price, ==1.0 for bank account   */

  /* the various "balances" are the sum of all of the values of 
   * all the splits in the account, up to and including this split */
  double  balance;
  double  cleared_balance;
  double  reconciled_balance;

  double  share_balance;
  double  share_cleared_balance;
  double  share_reconciled_balance;

} Split;


typedef struct _transaction {
  char  * num;               /* transaction id                            */
  Date    date;              /* transaction date                          */
  char  * description;        

  Split   credit_split;      /* credited account                          */
  Split   **debit_splits;    /* list of splits, null terminated           */

  char    write_flag;        /* used only during file IO                  */
} Transaction;


/** PROTOTYPES ******************************************************/

Split       * xaccMallocSplit (void);
void          xaccInitSplit   (Split *);    /* clears a split struct */
void          xaccFreeSplit   (Split *);    /* frees memory */
int           xaccCountSplits (Split **sarray);

Transaction * mallocTransaction (void);           /* mallocs and inits */
void          initTransaction (Transaction *);    /* clears a trans struct */

void          xaccTransSetDescription (Transaction *, char *);
void          xaccTransSetMemo (Transaction *, char *);
void          xaccTransSetAction (Transaction *, char *);
void          xaccTransSetReconcile (Transaction *, char);

void          xaccSplitSetMemo (Split *, char *);
void          xaccSplitSetAction (Split *, char *);
void          xaccSplitSetReconcile (Split *, char);

/* freeTransaction only does so if the transaction is not part of an
 * account. */
void          freeTransaction (Transaction *);

void xaccTransAppendSplit (Transaction *, Split *);
void xaccTransRemoveSplit (Transaction *, Split *);

/* The following two functions set the amount on the split */
void         xaccSetAmount (Split *, double);
void         xaccSetShareAmount (Split *, double);


/* The following four subroutines return the running balance up
 * to & including the indicated split.
 * 
 * The balance is the currency-denominated balance.  For accounts
 * with non-unit share prices, it is correctly adjusted for
 * share prices.
 * 
 * The share-balance is the number of shares. 
 * Price fluctuations do not change the share balance.
 * 
 * The cleared-balance is the currency-denominated balance 
 * of all transactions that have been marked as cleared or reconciled.
 * It is correctly adjusted for price fluctuations.
 * 
 * The reconciled-balance is the currency-denominated balance
 * of all transactions that have been marked as reconciled.
 */

double xaccGetBalance (Split *);
double xaccGetClearedBalance (Split *);
double xaccGetReconciledBalance (Split *);
double xaccGetShareBalance (Split *);

/********************************************************************\
 * sorting comparison function
 *
 * returns a negative value if transaction a is dated earlier than b,        
 * returns a positive value if transaction a is dated later than b,
 * returns zero if both transactions are on the same date.
 *
\********************************************************************/

int  xaccTransOrder (Transaction **ta, Transaction **tb);
int  xaccSplitOrder (Split **sa, Split **sb);

/*
 * count the number of transactions in the null-terminated array
 */
int xaccCountTransactions (Transaction **tarray);

/** GLOBALS *********************************************************/

#endif /* __XACC_TRANSACTION_H__ */
